# -*- coding: utf-8 -*

import os, sys, shutil, json, time
from posixpath import split
import platform
from . import IniParser
from Log.Logger import log2File
from pks import PksMgr
sys.path.append("./pycab")
import cabarchive as cab
from pycab.CabExtractor import CabExtractor
try:
    reload(sys)
    sys.setdefaultencoding('utf-8') # Python 2.x
except:
    pass

def get_brand(install_dir):
    json_file = os.path.join(install_dir, 'Data/asmodularize.dat')
    try:
        with open(json_file, 'r') as f:
            json_dict = json.load(f)
            for module in json_dict["module_info"]:
                if module.startswith("qaxav"):
                    return "qax"
    except:
        pass
    return str(2*180)

class FileMgr:
    def __init__(self, install_dir, update_data_dir, tmp_update_dir, tmp_updatecfg_conf, tmp_module_dat):
        self.install_dir = install_dir
        self.update_data_dir = update_data_dir
        self.tmp_update_dir = tmp_update_dir
        self.tmp_backup_dir = os.path.join(tmp_update_dir, 'backup')
        self.tmp_updatecfg_conf = tmp_updatecfg_conf
        self.tmp_module_dat = tmp_module_dat
        self.reboot = False
        self.isupdate = False
        self.lib_ext = self.get_lib_ext()
        self.pks_mgr = PksMgr.PksManager(install_dir)
        self.brand = get_brand(self.install_dir)

    def isExist(self, path):
        if os.path.exists(path):
            return True
        return False

    def copyToDir(self, srcFile, distDir):
        (filepath, filename) = os.path.split(srcFile)
        tempfilepath = distDir + filepath
        if os.path.exists(tempfilepath):
            shutil.rmtree(tempfilepath)
        os.makedirs(tempfilepath)
        shutil.copy(srcFile, tempfilepath)

    def setModuleIni(self, path_ini):

        iniParser = IniParser.IniParser(path_ini)
        iniParser.initconf()
        sections = iniParser.cf.sections()
        for section in sections:
            #for option in iniParser.cf.options(section):
            if iniParser.cf.has_option(section, 'path'):
                path = iniParser.get(section, "path")
                if path and 'productpath' in path:
                    new_path = path.replace("%productpath%", self.install_dir)
                    if not iniParser.set(section, 'path', new_path):
                        log2File(path_ini + " set productpath error")
                        return False
        return True

    def cmp_ver(self, ver1, ver2):
        lver1 = ver1.split('.')
        lver2 = ver2.split('.')
        f = 0
        if len(lver1) > len(lver2):
            f = len(lver1)
        else:
            f = len(lver2)
        for i in range(f):
            try:
                if int(lver1[i]) > int(lver2[i]):
                    return 1
                elif int(lver1[i]) == int(lver2[i]):
                    continue
                else:
                    return -1
            except IndexError as e:
                if len(lver1) > len(lver2):
                    return 1
                else:
                    return -1
        return 0

    def getModuleVer(self, module_name):
        mod_ver = ""
        with open(self.tmp_module_dat, 'r') as f:
            data = json.load(f)
            if not "module_info" in data:
                log2File(self.tmp_module_dat + ':' +  'module_info is not exist.')
            else:
                module_info = data["module_info"]
                for module in module_info:
                    if module == module_name:
                        mod_ver = data["module_info"][module_name]["ver"]
                        log2File("asmodulize.dat " + module_name + ' version is ' + mod_ver)
                        break
        return mod_ver

    def updateModuleDat(self, module_name, version):
        with open(self.tmp_module_dat, 'r') as f:
            data = json.load(f)
            if not "module_info" in data:
                log2File(self.tmp_module_dat + ':' +  'module_info is not exist.')
                return False
            else:
                module_info = data["module_info"]
                for module in module_info:
                    if module == module_name:
                        data["module_info"][module_name]["ver"] = version
                        data["module_info"][module_name]["time"] = time.strftime("%Y-%m-%d %H:%M:%S",time.localtime())
                        log2File("update asmodulize.dat " + module_name + ' version is ' + version)
                        break

        with open(self.tmp_module_dat, 'w') as f:
            #print data
            json.dump(data, f, indent=4)
        return True

    def __addWhiteAndVerify(self,module_name,file_list):
        #可信华泰的接口调用失败时返回模块升级失败
        #可信华泰的加白及验签文件每个模块默认都会有，且文件名是固定的
        ok = self.pks_mgr.HttcVerify(self.tmp_module_dir)
        if not ok:
            slog = ("update module %s failed in tmp_module_dir: %s," + 
                    "because HttcVerif failed") % (module_name,self.tmp_module_dir)
            log2File(slog)
            return False

        for file in file_list:
            tmp_file = os.path.join(self.tmp_module_dir, file)
            ok = self.pks_mgr.KySecUpdate(tmp_file)
            #一个失败就失败
            if not ok:
                break
        return ok

    def updateModuleImpl(self, module_name, path_ini):
        module_ini = IniParser.IniParser(path_ini)
        module_ini.initconf()
        update_ini = IniParser.IniParser(self.tmp_updatecfg_conf)
        update_ini.initconf()

        mod_ver = module_ini.get('qaxsafe', 'libver')
        if not mod_ver:
            mod_ver = module_ini.get('{}safe'.format(2*180), 'libver')
        if mod_ver == "":
            log2File("get module " + module_name + " version fail.")
            return False
        mod_ver_old = self.getModuleVer(module_name)
        log2File(module_name + ":new version is " + mod_ver + ", old version is " + mod_ver_old)
        if mod_ver_old == "":
            log2File( module_name + " version is empty, no need update")
            return True
        if self.cmp_ver(mod_ver, mod_ver_old) <=0:
            log2File( module_name + " is no need update")
            return True
        module_name_ext = module_name + self.lib_ext
        file_str = module_ini.get(module_name_ext, 'files0')
        log2File(module_name + '[' + file_str + ']')
        if file_str:
            file_list = file_str.split(',')
            #执行加白及验签动作,失败时直接返回
            ok = self.__addWhiteAndVerify(module_name,file_list)
            if not ok:
                return ok
            for file in file_list:
                tmp_file = os.path.join(self.tmp_module_dir, file)
                ver_new = module_ini.get(file, 'ver')
                md5_new = module_ini.get(file, 'md5')
                dir_new = module_ini.get(file, 'path')
                rname = module_ini.get(file, 'rname')
                if module_ini.cf.has_option(file, 'flag'):
                    flag = module_ini.get(file, 'flag')
                    if flag == "4":
                        self.reboot = True

                ver_old = ""
                if update_ini.cf.has_option(file.upper(), 'VER'):
                    ver_old = update_ini.get(file.upper(), 'VER')
                md5_old = ""
                if update_ini.cf.has_option(file.upper(), 'MD5'):
                    md5_old = update_ini.get(file.upper(), 'MD5')

                if not os.path.exists(dir_new):
                    os.makedirs(dir_new)
                file_new = os.path.join(dir_new, rname)
                if os.path.exists(file_new):
                    self.copyToDir(file_new, self.tmp_backup_dir)
                else:
                    fp = open(os.path.join(self.tmp_update_dir, 'new_update_file'),'a')
                    fp.write(file_new)
                    fp.write('\n')
                    fp.close()

                if ver_old == "" or self.cmp_ver(ver_new, ver_old) >0 or (self.cmp_ver(ver_new, ver_old) == 0 and md5_new != md5_old):
                    if os.path.exists(file_new):
                        os.remove(file_new)
                    shutil.copy(tmp_file, file_new)
                    log2File('update file:' + file_new + ', md5:' + md5_new)
                    self.isupdate = True
                    update_ini.set(file.upper(), 'VER', ver_new)
                    update_ini.set(file.upper(), 'PATH', file_new)
                    update_ini.set(file.upper(), 'MD5', md5_new)

        res = self.updateModuleDat(module_name, mod_ver)
        return res

    def __add_to_file(self, file_path, content):
        try:
            f = open(file_path, 'wb')
            # Adaptation python3 start
            if isinstance(content,str) and platform.python_version().startswith("3"):
                content = content.encode('UTF-8')
            # Adaptation python3 end
            f.write(content)
            f.close()
        except Exception as e:
            strInfo = 'add to file [%s] fail, content= [%s] ,err=%s' % (file_path, content, str(e))
            log2File(strInfo)
            if f:
                f.close()
            return False
        return True

    def __extractFiles(self, cab_path, file_path):
        if os.path.exists(cab_path):
            # Adaptation python3 start
            if platform.python_version().startswith("3"):
                try:
                    arc = cab.CabArchive()
                    with open(cab_path, "rb") as f:
                        old = f.read()
                        arc.parse(old)
                    cff = list(arc.values())[0]
                    filename = cff.filename
                    content = cff.contents
                    filepath = os.path.join(file_path, filename)
                    self.__add_to_file(filepath, content)
                except Exception as e:
                    strInfo = 'extract ' + cab_path + ' fail. ERROR[%s]' % str(e)
                    log2File(strInfo)
                    return False
            # Adaptation python3 end
            else:
                try:
                    arc = cab.CabArchive()
                    arc.parse_file(cab_path)
                    cff = arc.files[0]
                    filename = cff.filename
                    content = cff.contents
                    filepath = os.path.join(file_path, filename)
                    self.__add_to_file(filepath, content)
                except Exception as e:
                    strInfo = 'extract ' + cab_path + ' fail. ERROR[%s]' % str(e)
                    log2File(strInfo)
                    return False
        return True

    def extractFiles(self, path_module_dir, path_ini,module_name_ext):
        iniParser = IniParser.IniParser(path_ini)
        iniParser.initconf()
        sections = iniParser.cf.sections()
        files0 = iniParser.get(module_name_ext,"files0")
        files_list = files0.split(",")
        tmp_cab_dir = os.path.join(self.tmp_module_dir, "tmp_cab_dir")
        extractor = CabExtractor()
        sum_count = 0
        success_count = 0
        for section in files_list:
            try:
                url = iniParser.get(section, "url")
                if not url:
                    continue
                sum_count += 1
                cab_url_path = url.split("|")[0]
                modulePath = cab_url_path.split("/")[-2]
                cab_file_name = os.path.join(self.update_data_dir, modulePath, modulePath, os.path.basename(cab_url_path))
                if not os.path.exists(cab_file_name):
                    if modulePath == "engine_qax_linux_def":
                        modulePath = module_name_ext
                        cab_file_name = os.path.join(self.update_data_dir, modulePath, modulePath, os.path.basename(cab_url_path))

                if os.path.isdir(tmp_cab_dir):
                    shutil.rmtree(tmp_cab_dir)
                os.makedirs(tmp_cab_dir)
                self.__extractFiles(cab_file_name,tmp_cab_dir)
                src = os.path.join(tmp_cab_dir, os.listdir(tmp_cab_dir)[0])
                dst = os.path.join(self.tmp_module_dir, section)
                shutil.move(src, dst)
                success_count += 1
            except:
                log2File("cabextract error!"+cab_file_name)
                continue
        if os.path.isdir(tmp_cab_dir):
            shutil.rmtree(tmp_cab_dir)
        return False if success_count < sum_count else True

    def updateModule(self, module_name, module_dir):
        res = False
        module_name_ext = module_name + self.lib_ext
        module_ini = module_name_ext + '.ini'
        path_module_ini = os.path.join(module_dir, module_ini)
        path_module_dir = os.path.join(module_dir, module_name_ext)
        path_module_org = os.path.join(self.update_data_dir,module_name,module_name)
        if not self.isExist(path_module_ini):
            log2File(path_module_ini + " is not exits.")
            return res

        if not self.isExist(path_module_dir):
            log2File(path_module_dir + " is not exits.")
            if not self.isExist(path_module_org):
                log2File(path_module_org + " is not exits.")
                return res

        shutil.copy(path_module_ini, self.tmp_module_dir)
        tmp_module_ini = os.path.join(self.tmp_module_dir, module_ini)
        if not self.extractFiles(path_module_dir, tmp_module_ini,module_name_ext):
            return res
        if not self.setModuleIni(tmp_module_ini):
            return res
        if not self.updateModuleImpl(module_name, tmp_module_ini):
            return res
        res = True
        return res

    def endUpdate(self):
        if self.isupdate == False:
            return 0
        updatecfg_conf = os.path.join(self.install_dir, 'updatecfg.ini')
        if os.path.exists(updatecfg_conf):
            os.remove(updatecfg_conf)
            log2File('delete updatecfg.ini')
        shutil.copy(self.tmp_updatecfg_conf, updatecfg_conf)
        log2File('update updatecfg.ini')

        module_dat = os.path.join(self.install_dir, 'Data/asmodularize.dat')
        if os.path.exists(module_dat):
            os.remove(module_dat)
            log2File('delete asmodularize.dat')
        shutil.copy(self.tmp_module_dat, module_dat)
        log2File('update asmodularize.dat')

    def emptyTmpData(self):
        shutil.rmtree(self.tmp_update_dir)
        log2File('empty tmp update data directory')

    def update(self, module_name):
        res = False
        module_name_ext = module_name + self.lib_ext
        file_dir = os.path.join(self.update_data_dir, module_name_ext)
        org_module_dir = os.path.join(self.update_data_dir, module_name)
        if not self.isExist(file_dir):
            file_dir = os.path.join(self.update_data_dir, 'virus', module_name_ext)
            org_module_dir = os.path.join(self.update_data_dir, 'virus', module_name)
            if not self.isExist(file_dir):
                log2File(module_name_ext + ' is not exist.No need update')
                res = True
                return res
        self.tmp_module_dir = os.path.join(self.tmp_update_dir, module_name_ext)
        if os.path.exists(self.tmp_module_dir):
            shutil.rmtree(self.tmp_module_dir)
        os.mkdir(self.tmp_module_dir)

        if not self.updateModule(module_name, file_dir):
            return res
        res = True
        return res

    def rollback(self):
        log2File('roolback backup files')
        for (root, dirs, files) in os.walk(self.tmp_backup_dir):
            for filename in files:
                file_path = os.path.join(root, filename)
                real_path = file_path.replace(self.tmp_backup_dir, "")
                if os.path.exists(real_path):
                    os.remove(real_path)
                shutil.copy(file_path, real_path)
                log2File('rollback file ' + real_path)
        delete_path = os.path.join(self.tmp_update_dir, 'new_update_file')
        if os.path.exists(delete_path):
            with open(delete_path, 'r') as f:
                for line in f.readlines():
                    path = line.strip()
                    if os.path.exists(path):
                        os.remove(path)
                        log2File('rollback delete ' + path)

    def rebootService(self):
        if self.reboot and self.isupdate:
            brand = str(2*180)
            log2File("reboot {}safe service".format(brand))
            if self.install_dir == "/opt/{}safe".format(brand):
                os.system('cat /opt/{0}safe/.{0}entclient.pid| xargs kill -9'.format(brand))
            else:
                os.system('pgrep -f "start{}"|xargs kill -9'.format(self.brand))
                os.system('pgrep -f "{}safed"|xargs kill -9'.format(self.brand))
                p = os.popen('who | awk \'{print $1" "$2" "$5}\' | grep ":0" | awk \'{print $1}\' | head -1')
                user = p.read()
                user = user.strip()
                if user == "" or user == 'root':
                    log2File("canot reboot start{} in normal user, please start manually.".format(self.brand))
                else:
                    os.system('su ' + user + ' -c "nohup start{} >/dev/null 2>&1 &"'.format(self.brand))

    def get_lib_ext(self):
        lib_ext = ''
        update_conf = os.path.join(self.install_dir, 'conf/spec_info.conf')
        if os.path.exists(update_conf) == False:
            log2File('error:' + update_conf + 'not exist')
            return lib_ext

        with open(update_conf, 'r') as f:
            data = json.load(f)
            try:
                lib_ext = data['lib_ext']
            except  Exception as e:
                log2File('error:%s key value not found, error:%s' % (update_conf, str(e)))
        return lib_ext
if __name__ == "__main__":
    install_dir = "/opt/qaxsafe/"
    a = FileMgr(install_dir,'a', 'b', 'c', 'd')
    a.rebootService()
